import base64
import json
import logging
import threading
import queue
import hashlib
import secrets
from urllib.parse import urlencode
import webbrowser
import requests
import uuid
from flask import Flask, request, Response

# ---------------------------
# Здесь предполагается, что у вас есть реализации следующих функций в модуле utils:
# make_random_state() - генерирует случайное состояние (строку)
# write_to_config_file(content, filename) - сохраняет строку в файл (например, в папке конфигурации)
# read_from_config_file(filename) - читает строку из файла конфигурации
# Если их нет, замените их на свои реализации.
# ---------------------------
from .utils import make_random_state, write_to_config_file, read_from_config_file

# Константы (аналогичные Go‑константам)
CLIENT_ID = "com_jagex_auth_desktop_launcher"
STANDARD_LOGIN_CLIENT_ID = "1fddee4e-b100-4f4e-b2b0-097f9088f9d2"
LOGIN_PROVIDER = "runescape"
AUTH_URL = "https://account.jagex.com/oauth2/auth"
TOKEN_URL = "https://account.jagex.com/oauth2/token"
REDIRECT_URL = "https://secure.runescape.com/m=weblogin/launcher-redirect"
API_URL = "https://api.jagex.com/v1"
PROFILE_API_URL = "https://secure.jagex.com/rs-profile/v1"
SHIELD_URL = "https://auth.jagex.com/shield/oauth/token"
GAME_SESSION_API_URL = "https://auth.jagex.com/game-session/v1"
OSRS_BASIC_AUTH_HEADER = "Basic Y29tX2phZ2V4X2F1dGhfZGVza3RvcF9vc3JzOnB1YmxpYw=="


# Аналог класса-кортежа Tuple[T1, T2]
class Tuple:
    def __init__(self, first, second):
        self.first = first
        self.second = second


# ---------------------------
# Вспомогательные функции для PKCE
# ---------------------------
def make_random_verifier(length: int = 43) -> str:
    """
    Генерирует случайную строку для PKCE (code verifier)
    с использованием допустимых символов.
    """
    allowed_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~"
    result = []
    for _ in range(length):
        # Генерируем случайное 32-битное число, аналогично crypto.getRandomValues(new Uint32Array(43))
        rand_int = secrets.randbits(32)
        result.append(allowed_chars[rand_int % len(allowed_chars)])
    return "".join(result)


def compute_code_challenge(verifier: str) -> str:
    """
    Вычисляет code challenge по алгоритму S256:
      - SHA-256 от verifier,
      - Base64 URL-safe кодирование без завершающих '='.
    """
    verifier_bytes = verifier.encode("utf-8")
    digest = hashlib.sha256(verifier_bytes).digest()
    # base64.urlsafe_b64encode заменяет '+' на '-' и '/' на '_'
    code_challenge = base64.urlsafe_b64encode(digest).decode("utf-8").rstrip("=")
    return code_challenge


# ---------------------------
# OAuth2 конфигурация
# ---------------------------
def get_auth_config() -> dict:
    """
    Возвращает словарь с настройками OAuth2.
    """
    return {
        "client_id": CLIENT_ID,
        "client_secret": "",
        "scopes": ["openid", "offline", "gamesso.token.create", "user.profile.read"],
        "auth_url": AUTH_URL,
        "token_url": TOKEN_URL,
        "redirect_url": REDIRECT_URL,
    }


# ---------------------------
# Функция Login - инициирует OAuth2 авторизацию с PKCE
# ---------------------------
def login() -> None:
    """
    Инициирует OAuth2-авторизацию:
      – генерирует PKCE verifier и вычисляет code challenge,
      – генерирует случайное состояние,
      – сохраняет исходный verifier и состояние,
      – формирует URL авторизации и открывает его в браузере.
    """
    config = get_auth_config()
    verifier = make_random_verifier()  # Исходный verifier
    challenge = compute_code_challenge(
        verifier
    )  # Вычисляем code challenge по алгоритму S256
    state = make_random_state()

    params = {
        "client_id": config["client_id"],
        "redirect_uri": config["redirect_url"],
        "response_type": "code",
        "scope": " ".join(config["scopes"]),
        "state": state,
        "code_challenge": challenge,
        "code_challenge_method": "S256",
    }
    url = config["auth_url"] + "?" + urlencode(params)

    write_to_config_file(state, "state")
    # Сохраняем исходный verifier (не challenge!)
    write_to_config_file(verifier, "verifier")
    webbrowser.open(url)


def refresh_token(refresh_token: str):
    config = get_auth_config()
    data = {
        "client_id": config["client_id"],
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }
    response = requests.post(config["token_url"], data=data)
    if response.status_code != 200:
        raise Exception("Refresh token update failed: " + response.text)
    token = response.json()
    return token


# ---------------------------
# Функция Exchange - обмен authorization code на токен
# ---------------------------
def exchange(code: str) -> dict:
    """
    Обменивает authorization code на токен.
    Читает ранее сохранённый verifier и отправляет POST-запрос к TOKEN_URL.
    """
    config = get_auth_config()
    verifier = read_from_config_file("verifier")
    data = {
        "client_id": config["client_id"],
        "grant_type": "authorization_code",
        "code": code,
        "redirect_uri": config["redirect_url"],
        "code_verifier": verifier,
    }
    response = requests.post(config["token_url"], data=data)
    if response.status_code != 200:
        raise Exception("Token exchange failed: " + response.text)
    token = response.json()
    return token


# ---------------------------
# Функция ParseIdToken - разбирает JWT id_token
# ---------------------------
def parse_id_token(id_token: str) -> dict:
    """
    Разбирает JWT id_token:
      – декодирует заголовок и полезную нагрузку,
      – проверяет, что заголовок содержит тип 'JWT',
      – возвращает полезную нагрузку.
    """
    parts = id_token.split(".")
    if len(parts) != 3:
        raise Exception(f"malformed id_token: {len(parts)} sections, expected 3")

    def pad_base64(b: str) -> str:
        return b + "=" * (-len(b) % 4)

    # Декодирование заголовка
    header_bytes = base64.urlsafe_b64decode(pad_base64(parts[0]))
    header = json.loads(header_bytes)
    if header.get("typ") != "JWT":
        raise Exception(f"bad id_token header: typ {header.get('typ')}, expected JWT")

    # Декодирование полезной нагрузки
    payload_bytes = base64.urlsafe_b64decode(pad_base64(parts[1]))
    payload = json.loads(payload_bytes)
    logging.info("id_token payload: %s", payload)
    return payload


# ---------------------------
# Функция StandardLogin - стандартный процесс логина через Flask-сервер
# ---------------------------
def standard_login(id_token: str) -> tuple:
    """
    Реализует стандартный процесс логина:
      – меняет client_id, redirect_url и scopes,
      – формирует URL авторизации с дополнительными параметрами,
      – поднимает Flask-сервер для получения кода и нового id_token,
      – возвращает кортеж (code, id_token).
    """
    config = get_auth_config()
    config["client_id"] = STANDARD_LOGIN_CLIENT_ID
    config["redirect_url"] = "http://localhost"
    config["scopes"] = ["openid", "offline"]

    state = make_random_state()
    params = {
        "client_id": config["client_id"],
        "redirect_uri": config["redirect_url"],
        "response_type": "id_token code",
        "scope": " ".join(config["scopes"]),
        "state": state,
        "id_token_hint": id_token,
        "nonce": str(uuid.uuid4()),
        "prompt": "consent",
    }
    url = config["auth_url"] + "?" + urlencode(params)

    # Очередь для передачи результата из Flask в основной поток
    result_queue = queue.Queue()

    app = Flask(__name__)

    @app.route("/")
    def index():
        # Отправляем страницу с JavaScript, который извлекает фрагмент URL
        js = """
        <script>
            var fragment = window.location.hash.substring(1);
            var xhr = new XMLHttpRequest();
            xhr.open('GET', '/process_fragment?' + fragment, true);
            xhr.onreadystatechange = function() {
                if (xhr.readyState === XMLHttpRequest.DONE && xhr.status === 200) {
                    document.body.innerHTML = "You may now close this window";
                }
            };
            xhr.send();
        </script>
        """
        return Response(js, mimetype="text/html")

    @app.route("/process_fragment")
    def process_fragment():
        code = request.args.get("code")
        new_id_token = request.args.get("id_token")
        # print(code, new_id_token)
        result_queue.put((code, new_id_token))
        return "Received"

    def run_flask():
        port = 8080
        logging.info("Binding to port: %d", port)
        app.run(host="0.0.0.0", port=port, debug=False)

    flask_thread = threading.Thread(target=run_flask, daemon=True)
    flask_thread.start()

    webbrowser.open(url)

    # Ждем результата из очереди
    code, new_id_token = result_queue.get()
    # Здесь можно добавить остановку Flask-сервера, если требуется
    return code, new_id_token


# ---------------------------
# Функция GetGameSession - получение игровой сессии
# ---------------------------
def get_game_session(id_token: str) -> str:
    """
    Проверяет корректность id_token и отправляет POST-запрос к GAME_SESSION_API_URL/sessions для получения sessionId.
    """
    # Валидация id_token
    parse_id_token(id_token)

    url = GAME_SESSION_API_URL + "/sessions"
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Host": "auth.jagex.com",
    }
    body = {"idToken": id_token}
    response = requests.post(url, json=body, headers=headers)
    if response.status_code != 200:
        raise Exception("getGameSession failed: " + response.text)
    data = response.json()
    session_id = data.get("sessionId")
    if not isinstance(session_id, str):
        raise Exception(f"getGameSession: did not find 'sessionId', data: {data}")
    return session_id


# ---------------------------
# Функция GetUserDetails - получение данных пользователя
# ---------------------------
class UserDetails:
    def __init__(self, display_name: str, id: str, suffix: str, user_id: str):
        self.display_name = display_name
        self.id = id
        self.suffix = suffix
        self.user_id = user_id


def get_user_details(sub: str, access_token: str) -> UserDetails:
    """
    Получает данные пользователя по sub, отправляя GET-запрос к API.
    """
    url = f"{API_URL}/users/{sub}/displayName"
    headers = {"Authorization": "Bearer " + access_token}
    response = requests.get(url, headers=headers)
    if response.status_code != 200:
        raise Exception("could not get user details: " + response.text)
    info = response.json()
    return UserDetails(
        display_name=info.get("displayName"),
        id=info.get("id"),
        suffix=info.get("suffix"),
        user_id=info.get("userId"),
    )


# ---------------------------
# Функция GetAccounts - получение списка аккаунтов
# ---------------------------
class Account:
    def __init__(self, account_id: str, display_name: str, user_hash: str):
        self.account_id = account_id
        self.display_name = display_name
        self.user_hash = user_hash


def get_accounts(session_id: str) -> list:
    """
    Получает список аккаунтов, отправляя GET-запрос к GAME_SESSION_API_URL/accounts.
    """
    url = f"{GAME_SESSION_API_URL}/accounts"
    headers = {
        "Accept": "application/json",
        "Authorization": "Bearer " + session_id,
    }
    response = requests.get(url, headers=headers)
    if response.status_code != 200:
        raise Exception("getAccounts error: " + response.text)
    accounts_json = response.json()
    accounts = []
    for acc in accounts_json:
        accounts.append(
            Account(
                account_id=acc.get("accountId"),
                display_name=acc.get("displayName"),
                user_hash=acc.get("userHash"),
            )
        )
    return accounts


# ---------------------------
# Функция getShieldTokens - обмен токена на shield-токен
# ---------------------------
def get_shield_tokens(access_token: str) -> dict:
    """
    Отправляет запрос на обмен токена (token_exchange) к SHIELD_URL.
    """
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": OSRS_BASIC_AUTH_HEADER,
    }
    params = {
        "token": access_token,
        "grant_type": "token_exchange",
        "scope": "gamesso.token.create",
    }
    response = requests.post(SHIELD_URL, data=params, headers=headers)
    if response.status_code != 200:
        raise Exception("getShieldTokens failed: " + response.text)
    data = response.json()
    logging.info("getShieldToken: %s", response.text)
    return data
