import curses
from typing import List


class NoAccountChosenError(Exception):
    def __str__(self):
        return "no account was chosen"


# Пример класса для аккаунта (адаптируйте под вашу реализацию)
class Account:
    def __init__(self, account_id: str, display_name: str):
        self.account_id = account_id
        self.display_name = display_name


# Пример класса для данных пользователя
class UserDetails:
    def __init__(self, display_name: str, suffix: str):
        self.display_name = display_name
        self.suffix = suffix


def curses_account_selection(
    stdscr, user: UserDetails, accounts: List[Account]
) -> Account:
    # Настройка curses
    curses.curs_set(0)  # скрываем курсор
    stdscr.nodelay(False)
    stdscr.keypad(True)

    # Инициализация цветовой пары (белый текст на синем фоне)
    curses.start_color()
    curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_CYAN)
    highlight_color = curses.color_pair(1)

    current_index = 0
    num_accounts = len(accounts)

    while True:
        stdscr.clear()
        height, width = stdscr.getmaxyx()

        # Заголовок с данными пользователя
        title = f"You are logged in as: {user.display_name}#{user.suffix}"
        stdscr.addstr(1, 2, title)

        # Инструкция
        instruction = (
            "Use UP/DOWN to select an account, Enter to confirm, 'q' to cancel."
        )
        stdscr.addstr(3, 2, instruction)

        # Вывод списка аккаунтов
        for idx, acc in enumerate(accounts):
            line = f"{idx + 1}. {acc.display_name}"
            y = 5 + idx
            x = 4
            if idx == current_index:
                stdscr.attron(highlight_color)
                stdscr.addstr(y, x, line)
                stdscr.attroff(highlight_color)
            else:
                stdscr.addstr(y, x, line)

        stdscr.refresh()

        key = stdscr.getch()

        if key in [curses.KEY_UP, ord("k")]:
            current_index = (current_index - 1) % num_accounts
        elif key in [curses.KEY_DOWN, ord("j")]:
            current_index = (current_index + 1) % num_accounts
        elif key in [10, 13]:  # Enter
            return accounts[current_index]
        elif key in [ord("q"), 27]:  # q или Esc для выхода
            raise NoAccountChosenError()


def get_chosen_account(user: UserDetails, accounts: List[Account]) -> Account:
    """
    Запускает TUI для выбора аккаунта.

    :param user: Объект с информацией о пользователе (например, display_name и suffix)
    :param accounts: Список объектов Account
    :return: Выбранный объект Account
    :raises NoAccountChosenError: Если выбор не был произведён
    """
    try:
        chosen = curses.wrapper(curses_account_selection, user, accounts)
        return chosen
    except Exception as e:
        raise NoAccountChosenError() from e


# Пример использования:
if __name__ == "__main__":
    # Создаем тестовые данные
    user = UserDetails("PlayerOne", "1234")
    accounts = [
        Account("acc1", "Warrior"),
        Account("acc2", "Mage"),
        Account("acc3", "Archer"),
    ]

    try:
        chosen_account = get_chosen_account(user, accounts)
        print(
            f"Selected account: {chosen_account.display_name} (ID: {chosen_account.account_id})"
        )
    except NoAccountChosenError as e:
        print(str(e))
