import os
import sys
import json
import logging
import random
import string
from typing import Any, Dict, NamedTuple


# Для удобства можно определить кортеж, аналог Tuple[T1, T2] в Go.
class Tuple(NamedTuple):
    first: Any
    second: Any


def make_random_state() -> str:
    """
    Генерирует случайную строку длиной 12 символов,
    состоящую из латинских букв (заглавных и строчных).
    """
    charset = string.ascii_letters
    length = 12
    return "".join(random.choice(charset) for _ in range(length))


def get_config_file(filename: str) -> str:
    """
    Формирует путь до файла конфигурации, расположенного в пользовательской папке.
    По умолчанию используется директория "~/.config/pyscape".
    """
    # Получаем директорию конфигурации пользователя.
    config_dir = os.path.expanduser("~/.config")
    return os.path.join(config_dir, "pyscape", filename)


def write_to_config_file(content: str, filename: str) -> None:
    """
    Записывает content в файл конфигурации, создавая необходимые каталоги.
    """
    config_file = get_config_file(filename)
    os.makedirs(os.path.dirname(config_file), exist_ok=True)
    print("content: ", content)
    try:
        with open(config_file, "w", encoding="utf-8") as f:
            f.write(content)
    except Exception as e:
        print(str(e))
        raise


def read_from_config_file(filename: str) -> str:
    """
    Читает и возвращает содержимое файла конфигурации.
    """
    config_file = get_config_file(filename)
    with open(config_file, "r", encoding="utf-8") as f:
        contents = f.read()
    return contents


def parse_intent_payload(payload: str) -> Dict[str, str]:
    """
    Разбирает строку payload формата "jagex:key1=value1,key2=value2"
    и возвращает словарь с полученными парами ключ-значение.
    """
    result: Dict[str, str] = {}
    parts = payload.split(":")
    if parts[0] != "jagex" or len(parts) != 2:
        logging.fatal("invalid payload: %s", payload)
        sys.exit(1)
    pairs = parts[1].split(",")
    for pair in pairs:
        key_value = pair.split("=")
        if len(key_value) == 2:
            result[key_value[0]] = key_value[1]
    return result


def parse_json(buffer: bytes) -> Dict[str, Any]:
    """
    Десериализует JSON-данные из байтов в словарь.
    """
    return json.loads(buffer.decode("utf-8"))


def die(err: Exception) -> None:
    """
    Выводит ошибку, ждёт нажатия Enter и завершает работу программы.
    """
    logging.error(err)
    print("Press Enter to exit...")
    input()
    sys.exit(1)
